/*
 * All the boards in the library
 *
 * Copyright (C) 2013 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2 as published by the Free Software Foundation or, at your
 * option, any later version.
 */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include "adc-lib.h"
#include "adc-lib-int.h"

#define ADC_NAME_FMCADC100M "fmc-adc-100m14b4cha"
#define ADC_NAME_ZIOFAKE "adc-zzero"
#define ADC_NAME_GENERICFAKE "adc-genericfake"

/**
 * List of supported board names
 */
const char *adc_board_name[] = {
#ifdef SUPPORT_FMCADC_100MS_4CH_14BIT
	[FMCADC_100MS_4CH_14BIT] = ADC_NAME_FMCADC100M,
#endif
	[ADC_ZIOFAKE] = ADC_NAME_ZIOFAKE,
	[ADC_GENERICFAKE] = ADC_NAME_GENERICFAKE,
};


#ifdef SUPPORT_FMCADC_100MS_4CH_14BIT
#define ADC_100M_4CH_14BIT_ACQ_MASK (1LL << ADC_CONF_ACQ_N_SHOTS) | \
				    (1LL << ADC_CONF_ACQ_POST_SAMP) | \
				    (1LL << ADC_CONF_ACQ_PRE_SAMP) | \
				    (1LL << ADC_CONF_ACQ_UNDERSAMPLE) | \
				    (1LL << ADC_CONF_ACQ_FREQ_HZ) | \
				    (1LL << ADC_CONF_ACQ_N_BITS)
#define ADC_100M_4CH_14BIT_CHN_MASK (1LL << ADC_CONF_CHN_RANGE) | \
				    (1LL << ADC_CONF_CHN_TERMINATION) | \
				    (1LL << ADC_CONF_CHN_OFFSET) | \
				    (1LL << ADC_CONF_CHN_SATURATION)
#define ADC_100M_4CH_14BIT_BRD_MASK (1LL << ADC_CONF_BRD_STATE_MACHINE_STATUS) | \
				    (1LL << ADC_CONF_BRD_N_CHAN) | \
				    (1LL << ADC_CONF_BRD_N_TRG_EXT) | \
				    (1LL << ADC_CONF_BRD_N_TRG_THR) | \
				    (1LL << ADC_CONF_BRD_N_TRG_TIM)
#define ADC_100M_4CH_14BIT_TRG_EXT_MASK (1LL << ADC_CONF_TRG_EXT_ENABLE) | \
					(1LL << ADC_CONF_TRG_EXT_POLARITY) | \
					(1LL << ADC_CONF_TRG_EXT_DELAY)
#define ADC_100M_4CH_14BIT_TRG_THR_MASK (1LL << ADC_CONF_TRG_THR_ENABLE) | \
					(1LL << ADC_CONF_TRG_THR_POLARITY) | \
					(1LL << ADC_CONF_TRG_THR_DELAY) | \
					(1LL << ADC_CONF_TRG_THR_THRESHOLD) | \
					(1LL << ADC_CONF_TRG_THR_HYSTERESIS)
#define ADC_100M_4CH_14BIT_TRG_TIM_MASK 0x0

struct adc_operations fa_100ms_4ch_14bit_op = {
	.open =		adc_100m14b4cha_open,
	.close =		adc_zio_close,

	.acq_start =		adc_100m14b4cha_acq_start,
	.acq_poll =		adc_zio_acq_poll,
	.acq_stop =		adc_100m14b4cha_acq_stop,

	.apply_config =	adc_100m14b4cha_apply_config,
	.retrieve_config =	adc_100m14b4cha_retrieve_config,

	.get_param =		adc_zio_get_param,
	.set_param =		adc_zio_set_param,

	.request_buffer =	adc_zio_request_buffer,
	.fill_buffer =		adc_zio_fill_buffer,
	.tstamp_buffer =	adc_zio_tstamp_buffer,
	.release_buffer =	adc_zio_release_buffer,
	.trigger_fire =	adc_100m14b4cha_trigger_fire,
};

struct adc_board_type fmcadc_100ms_4ch_14bit = {
	.name = ADC_NAME_FMCADC100M,	/* for library open() */
	.devname = "adc-100m14b",	/* for device named in /dev/zio */
	.driver_type = "zio",
	.capabilities = {
		[ADC_CONF_TYPE_BRD] = ADC_100M_4CH_14BIT_BRD_MASK,
		[ADC_CONF_TYPE_CHN] = ADC_100M_4CH_14BIT_CHN_MASK,
		[ADC_CONF_TYPE_ACQ] = ADC_100M_4CH_14BIT_ACQ_MASK,
		[ADC_CONF_TYPE_TRG_EXT] = ADC_100M_4CH_14BIT_TRG_EXT_MASK,
		[ADC_CONF_TYPE_TRG_THR] = ADC_100M_4CH_14BIT_TRG_THR_MASK,
		[ADC_CONF_TYPE_TRG_TIM] = ADC_100M_4CH_14BIT_TRG_TIM_MASK,
	},
	.adc_op = &fa_100ms_4ch_14bit_op,
};
#endif


#define ADC_ZIO_ACQ_MASK (1LL << ADC_CONF_ACQ_N_SHOTS) | \
			 (1LL << ADC_CONF_ACQ_POST_SAMP) | \
			 (1LL << ADC_CONF_ACQ_PRE_SAMP) | \
			 (1LL << ADC_CONF_ACQ_N_BITS)
#define ADC_ZIO_CHN_MASK (1LL << ADC_CONF_CHN_OFFSET) | \
			 (1LL << ADC_CONF_CHN_GAIN)
#define ADC_ZIO_BRD_MASK (1LL << ADC_CONF_BRD_N_CHAN)


struct adc_operations fa_zio_fake_op = {
	.open = 		adc_ziofake_open,
	.close =		adc_zio_close,

	.acq_start =		adc_zio_acq_start,
	.acq_poll =		adc_zio_acq_poll,
	.acq_stop =		adc_zio_acq_stop,

	.apply_config = 	adc_zio_apply_config,
	.retrieve_config =	adc_zio_retrieve_config,

	.get_param =		adc_zio_get_param,
	.set_param =		adc_zio_set_param,

	.request_buffer =	adc_zio_request_buffer,
	.fill_buffer =		adc_zio_fill_buffer,
	.tstamp_buffer =	adc_zio_tstamp_buffer,
	.release_buffer =	adc_zio_release_buffer,
};

struct adc_board_type adc_ziofake = {
	.name = ADC_NAME_ZIOFAKE,
	.devname = "zzero",
	.driver_type = "zio",
	.capabilities = {
		[ADC_CONF_TYPE_BRD] = ADC_ZIO_BRD_MASK,
		[ADC_CONF_TYPE_CHN] = ADC_ZIO_CHN_MASK,
		[ADC_CONF_TYPE_ACQ] = ADC_ZIO_ACQ_MASK,
		[ADC_CONF_TYPE_TRG_EXT] = 0x0,
		[ADC_CONF_TYPE_TRG_THR] = 0x0,
		[ADC_CONF_TYPE_TRG_TIM] = 0x0,
	},
	.adc_op = &fa_zio_fake_op,
};



struct adc_operations fa_generic_fake_op = {
	.open = 		adc_genfake_open,
	.close =		adc_genfake_close,

	.acq_start =		adc_genfake_acq_start,
	.acq_poll = 		adc_genfake_acq_poll,
	.acq_stop =		adc_genfake_acq_stop,

	.apply_config = 	adc_genfake_apply_config,
	.retrieve_config =	adc_genfake_retrieve_config,

	.set_param = 		adc_genfake_set_param,
	.get_param =		adc_genfake_get_param,

	.request_buffer =	adc_genfake_request_buffer,
	.fill_buffer =		adc_genfake_fill_buffer,
	.tstamp_buffer =	adc_genfake_tstamp_buffer,
	.release_buffer =	adc_genfake_release_buffer,
};

#define ADC_GENERICFAKE_BRD_MASK (1LL << ADC_CONF_BRD_N_CHAN)

#define ADC_GENERICFAKE_ACQ_MASK (1LL << ADC_CONF_ACQ_N_SHOTS) | \
	(1LL << ADC_CONF_ACQ_POST_SAMP) |			 \
	(1LL << ADC_CONF_ACQ_PRE_SAMP) | \
	(1LL << ADC_CONF_ACQ_N_BITS)
struct adc_board_type adc_genericfake = {
	.name = ADC_NAME_GENERICFAKE,
	.devname = "fakedev",
	.driver_type = "fakebus",
	.capabilities = {
		[ADC_CONF_TYPE_BRD] = ADC_GENERICFAKE_BRD_MASK,
		[ADC_CONF_TYPE_CHN] = 0x0,
		[ADC_CONF_TYPE_ACQ] = ADC_GENERICFAKE_ACQ_MASK,
		[ADC_CONF_TYPE_TRG_EXT] = 0x0,
		[ADC_CONF_TYPE_TRG_EXT] = 0x0,
		[ADC_CONF_TYPE_TRG_THR] = 0x0,
		[ADC_CONF_TYPE_TRG_TIM] = 0x0,
	},
	.adc_op = &fa_generic_fake_op,
};

/*
 * The following array is the main entry point into the boards
 */
static const struct adc_board_type *adc_board_types[] = {
#ifdef SUPPORT_FMCADC_100MS_4CH_14BIT
	[FMCADC_100MS_4CH_14BIT] = &fmcadc_100ms_4ch_14bit,
#endif
	[ADC_ZIOFAKE] = &adc_ziofake,
	[ADC_GENERICFAKE] = &adc_genericfake,
	/* add new boards here */
};


static const struct adc_board_type *find_board(char *name)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(adc_board_types); i++)
		if (adc_board_types[i] && !strcmp(name, adc_board_types[i]->name))
			return adc_board_types[i];
	errno = ENODEV;
	return NULL;
}


/**
 * It opens a link with an ADC device
 * @param[in] name name of the board type to open according to adc_board_name.
 * @param[in] dev_id The device identifier is used to enumerate several
 *     cards in the same system.  The number is usually dependent on the
 *     geographic placement of the card (bus number, slot number).
 * @param[in] totalsamples This is a a hint about how big a buffer the
 *     application will use. i.e. ``nshots * (presamples + postsamples)``
 * @param[in] nbuffer This is a hint about how many buffers are being used
 *     at the same time by the application. For example, a multi-shot
 *     acquisition requires all buffers to be available at the same time.
 *     Again, this information may require device-level configuration at
 *     open time.
 * @param[in] flags currently it supports: ADC_F_FLUSH, ADC_F_VERBOSE
 *
 * @return It returns an opaque object describing the device.
 *         It should be passed around but not be looked into.
 *
 * @todo Open should choose the buffer type
 */
struct adc_dev *adc_open(char *name, unsigned int dev_id,
			 unsigned long totalsamples,
			 unsigned int nbuffer,
			 unsigned long flags)
{
	const struct adc_board_type *b;

	b = find_board(name);
	if (!b)
		return NULL;

	return b->adc_op->open(b, dev_id, totalsamples, nbuffer, flags);
}

#define ADC_PATH_PATTERN "/dev/%s.%d"

/**
 * It opens a link with an ADC device using a unit number identifier
 * @param[in] name name of the board type to open according to adc_board_name.
 * @param[in] lun the Logical Unit Number. This number comes from a database
 *     description of the system.
 * @param[in] buffersize This is a a hint about how big a buffer the
 *     application will use. i.e. ``nshots * (presamples + postsamples)``
 * @param[in] nbuffer This is a hint about how many buffers are being used
 *     at the same time by the application.
 * @param[in] flags currently it supports: ADC_F_FLUSH, ADC_F_VERBOSE
 *
 * @return It returns an opaque object describing the device.
 *         It should be passed around but not be looked into.
 *
 * @todo make it generic, now it works only for the CERN environment
 *       with the adc-100m14b. Actually I (Federico) believe that this
 *       is impossible to make it generic enough
 * @todo some udev rule should create symlink to map lun->devid
 */
struct adc_dev *adc_open_by_lun(char *name, int lun,
				unsigned long buffersize,
				unsigned int nbuffer,
				unsigned long flags)
{
	ssize_t ret;
	char dev_id_str[8];
	char path[PATH_MAX];
	int dev_id;

	ret = snprintf(path, sizeof(path), "/dev/%s.%d",
		       "adc-100m14b" /* FIXME: this must be generic */,
		       lun);
	if (ret < 0 || ret >= sizeof(path)) {
		errno = EINVAL;
		return NULL;
	}
	ret = readlink(path, dev_id_str, sizeof(dev_id_str));
	if (sscanf(dev_id_str, "%4x", &dev_id) != 1) {
		errno = ENODEV;
		return NULL;
	}
	return adc_open(name, dev_id, buffersize, nbuffer, flags);

}


/**
 * It closes the link with the ADC devices previously opened with adc_open()
 * @param[in] dev the device token from adc_open()
 * @return It returns 0 on success, otherwise -1 and errno is appropriately set
 */
int adc_close(struct adc_dev *dev)
{
	struct adc_gid *b = (struct adc_gid *)dev;

	return b->board->adc_op->close(dev);
}
